<?php
/**
 * The Whups_Ticket:: class encapsulates some logic relating to
 * tickets, sending updates, etc.
 *
 * $Horde: whups/lib/Ticket.php,v 1.27 2005/01/12 17:04:45 chuck Exp $
 *
 * Copyright 2004-2005 Chuck Hagenbuch <chuck@horde.org>
 *
 * See the enclosed file LICENSE for license information (BSD). If you
 * did not receive this file, see http://www.horde.org/licenses/bsdl.php.
 *
 * @author  Chuck Hagenbuch <chuck@horde.org>
 * @package Whups
 */
class Whups_Ticket {

    /**
     * The id of the ticket this object wraps.
     *
     * @var integer $_id
     */
    var $_id;

    /**
     * The current values of the ticket.
     *
     * @var array $_details
     */
    var $_details;

    /**
     * Array of changes to make to the ticket.
     *
     * @var array $_changes
     */
    var $_changes = array();

    /**
     * Load the ticket info for $id, returning either a Whups_Ticket
     * object or a PEAR_Error.
     *
     * @static
     *
     * @param integer $id  The ticket id.
     *
     * @return object PEAR_Error | object Whups_Ticket
     */
    function &makeTicket($id)
    {
        global $whups;

        $details = $whups->getTicketDetails($id);
        if (is_a($details, 'PEAR_Error')) {
            return $details;
        } else {
            return $ticket = &new Whups_Ticket($id, $details);
        }
    }

    /**
     * Create a new ticket. Pretty bare wrapper around
     * Whups_Driver::addTicket().
     *
     * @param array $info  The ticket create form data.
     *
     * @return object  Either a Whups_Ticket object on success or
     *                 a PEAR_Error object on failure.
     */
    function &newTicket($info, $requester)
    {
        $id = $GLOBALS['whups']->addTicket($info, $requester);
        if (is_a($id, 'PEAR_Error')) {
            return $id;
        } else {
            $ticket = &Whups_Ticket::makeTicket($id);
            $ticket->notify($ticket->get('user_id_requester'), true);
            return $ticket;
        }
    }

    /**
     * Constructs a new Whups_Ticket option from an array of
     * information (as returned by the backend).
     *
     * @param integer $id       The ticket id.
     * @param array   $details  The array of ticket information.
     */
    function Whups_Ticket($id, $details)
    {
        $this->_id = $id;
        $this->_details = $details;
    }

    /**
     * Get back an array of all ticket information.
     *
     * @return array  All ticket information.
     */
    function getDetails()
    {
        return $this->_details;
    }

    /**
     * Return the current ticket's id.
     *
     * @return integer  Ticket id.
     */
    function getId()
    {
        return $this->_id;
    }

    /**
     * Get a piece of information from this ticket.
     *
     * @param string $detail  The detail to get.
     *
     * @return mixed  The detail value.
     */
    function get($detail)
    {
        return isset($this->_details[$detail]) ? $this->_details[$detail] : null;
    }

    /**
     * Immediately changes an attribute of the ticket to a new
     * value. Never touches the backend; do not use for changes that
     * you want to persist.
     */
    function set($detail, $value)
    {
        $this->_details[$detail] = $value;
    }

    /**
     * Notes that an attribute of the ticket should change, but does
     * not actually make the change until commit() is called.
     *
     * @see commit()
     *
     * @var string $detail  The attribute to change.
     * @var string $value   The new value of $detail.
     */
    function change($detail, $value)
    {
        if (!isset($this->_details[$detail]) || $this->_details[$detail] != $value) {
            $this->_changes[$detail] = array('from' => $this->get($detail),
                                             'from_name' => $this->get($detail . '_name'),
                                             'to' => $value);
        }
    }

    /**
     * Go through a list of built-up changes and commit them to the
     * backend. This will send email updates, update the ticket log,
     * etc.
     *
     * @see change()
     *
     * @return boolean | object PEAR_Error
     */
    function commit($author = null)
    {
        global $whups;

        if (!count($this->_changes)) {
            return true;
        }

        if (is_null($author)) {
            $author = Auth::getAuth();
        }

        // Get a new transaction id from the backend.
        $transaction = $whups->newTransaction();
        if (is_a($transaction, 'PEAR_Error')) {
            return $transaction;
        }

        $updates = array();
        foreach ($this->_changes as $detail => $values) {
            $value = $values['to'];
            switch ($detail) {
            case 'owners':
                // Fetch $oldOwners list; then loop through $value
                // adding and deleting as needed.
                $oldOwners = $whups->getOwners($this->_id);
                $this->_changes['oldowners'] = $oldOwners;

                foreach ($value as $owner) {
                    if (empty($oldOwners[$owner])) {
                        $whups->addTicketOwner($this->_id, $owner);
                        $whups->updateLog($this->_id, $author, array('assign' => $owner), $transaction);
                    } else {
                        // Remove $owner from the old owners list;
                        // anyone left in $oldOwners will be removed.
                        unset($oldOwners[$owner]);
                    }
                }

                // Delete removed owners and log the removals.
                if (is_array($oldOwners)) {
                    foreach ($oldOwners as $owner) {
                        $whups->deleteTicketOwner($this->_id, $owner);
                        $whups->updateLog($this->_id, $author, array('unassign' => $owner), $transaction);
                    }
                }
                break;

            case 'comment':
                $comment_email = isset($this->_changes['comment-email']['to']) ? $this->_changes['comment-email']['to'] : null;
                $commentId = $whups->addComment($this->_id, $value, $author, $comment_email);
                if (is_a($commentId, 'PEAR_Error')) {
                    return $commentId;
                } else {
                    // Store the comment id in the updates array for
                    // the log.
                    $updates['comment'] = $commentId;
                    if (!empty($this->_changes['comment-perms'])) {
                        $this->addCommentPerms($commentId, $this->_changes['comment-perms']['to']);
                    }

                    // If this is a guest update, the comment id is
                    // going to map to the requester pseudo-username.
                    if ($author === false) {
                        $author = $creator = '-' . $commentId . '_comment';
                    }
                }
                break;

            case 'comment-email':
            case 'comment-perms':
                // Skip these, handled in the comment case.
                break;

            case 'attachment':
                $result = $this->addAttachment($value['name'], $value['tmp_name']);
                if (is_a($result, 'PEAR_Error')) {
                    return $result;
                } else {
                    // Store the new file name in the updates array
                    // for the log.
                    $updates['attachment'] = $value['name'];
                }
                break;

            default:
                $updates[$detail] = $value;
            }
        }

        if (count($updates)) {
            $result = $whups->updateTicket($this->_id, $updates);
            if (is_a($result, 'PEAR_Error')) {
                return $result;
            }

            $result = $whups->updateLog($this->_id, $author, $updates, $transaction);
            if (is_a($result, 'PEAR_Error')) {
                return $result;
            }
        }

        // Reload $this->_details to make sure we have the latest
        // information.
        //
        // @TODO Only touch the db if we have to.
        $details = $whups->getTicketDetails($this->_id);
        if (is_a($details, 'PEAR_Error')) {
            return $details;
        } else {
            $this->_details = array_merge($this->_details, $details);
        }

        // Send notification emails to all ticket listeners.
        $this->notify($author, false);

        // Reset the changes array.
        $this->_changes = array();

        return true;
    }

    /**
     * Adds an attachment to this ticket.
     */
    function addAttachment(&$attachment_name, $attachment_file)
    {
        global $conf;

        if (!isset($conf['vfs']['type'])) {
            return PEAR::raiseError(_("The VFS backend needs to be configured to enable attachment uploads."), 'horde.error');
        }

        require_once 'VFS.php';
        $vfs = &VFS::singleton($conf['vfs']['type'], Horde::getDriverConfig('vfs'));
        if (is_a($vfs, 'PEAR_Error')) {
            return $vfs;
        }

        $dir = WHUPS_VFS_ATTACH_PATH . '/' . $this->_id;
        while ($vfs->exists($dir, $attachment_name)) {
            if (preg_match('/(.*)\[(\d+)\](\.[^.]*)?$/', $attachment_name, $match)) {
                $attachment_name = $match[1] . '[' . ++$match[2] . ']' . $match[3];
            } else {
                $dot = strrpos($attachment_name, '.');
                if ($dot === false) {
                    $attachment_name .= '[1]';
                } else {
                    $attachment_name = substr($attachment_name, 0, $dot) . '[1]' . substr($attachment_name, $dot);
                }
            }
        }

        return $vfs->write($dir, $attachment_name, $attachment_file, true);
    }

    /**
     * Redirect to the specified view of this ticket.
     *
     * @param optional string $view  What ticket screen to show.
     */
    function show($view = '')
    {
        header('Location: ' . Horde::applicationUrl(Util::addParameter('ticket/', 'id', $this->_id), true));
        exit;
    }

    /**
     * Set exclusive read permissions on a comment to $group.
     *
     * @param integer $commentId  The id of the comment to restrict.
     * @param string  $group      The group name to limit access by.
     */
    function addCommentPerms($commentId, $group)
    {
        if (!empty($group)) {
            $perm = &$GLOBALS['perms']->newPermission('whups:comments:' . $commentId);
            $perm->addGroupPermission($group, PERMS_READ, false);
            return $GLOBALS['perms']->addPermission($perm);
        }
    }

    /**
     * Set all properties of the ticket necessary to display the
     * TicketDetailsForm.
     *
     * @param object Horde_Variables $vars  The form variables object to set info in.
     */
    function setDetails(&$vars)
    {
        $vars->set('id', $this->getId());
        foreach ($this->getDetails() as $varname => $value) {
            $vars->set($varname, $value);
        }

        /* User formatting. */
        $vars->set('user_id_requester', Whups::formatUser($this->get('user_id_requester')));
        $vars->set('user_id_owner', Whups::getOwners($this->_id));

        /* Attachments. */
        $attachments = array();
        $files = Whups::getAttachments($this->_id);
        if ($files) {
            if (is_a($files, 'PEAR_Error')) {
                $notification->push($files);
            } else {
                foreach ($files as $file) {
                    $attachments[] = Whups::attachmentUrl($this->_id, $file);
                }
            }
            $vars->set('attachments', implode("<br />\n", $attachments));
        }
    }

    /**
     * Notify all appropriate people of the creation/update of this ticket.
     *
     * @param string  $author  Who created/changed the ticket?
     * @param boolean $isNew   Is this a new ticket or a change to an existing
     *                         one?
     */
    function notify($author, $isNew)
    {
        global $conf, $whups;

        $fields = array('queue' => _("Queue"),
                        'version' => _("Version"),
                        'state' => _("State"),
                        'priority' => _("Priority"),
                        'type' => _("Type"));

        /* Find longest translated field name. */
        $length = 0;
        foreach (array_merge($fields, array(_("Created By"),
                                            _("Updated By"),
                                            _("Summary"),
                                            _("Owners"),
                                            _("Attachment"))) as $field) {
            $length = max($length, String::length($field));
        }

        /* Build up the ticket message. URL first. */
        $message = sprintf(_("Ticket URL: %s\n"), Util::addParameter(Horde::applicationUrl('ticket/', true, -1), 'id', $this->_id));

        /* Ticket properties. */
        $message .= "-----------------------------------------------------------------------\n";
        $message .= ' ' . String::pad(_("Ticket"), $length) . ' | ' . $this->_id . "\n";
        $message .= ' ' . String::pad($isNew ? _("Created By") : _("Updated By"), $length) . ' | ' . Whups::formatUser($author) . "\n";
        if (isset($this->_changes['summary'])) {
            $message .= '-' . String::pad(_("Summary"), $length) . ' | ' . $this->_changes['summary']['from'] . "\n";
            $message .= '+' . String::pad(_("Summary"), $length) . ' | ' . $this->get('summary') . "\n";
        } else {
            $message .= ' ' . String::pad(_("Summary"), $length) . ' | ' . $this->get('summary') . "\n";
        }

        $fields = array('queue' => _("Queue"),
                        'version' => _("Version"),
                        'state' => _("State"),
                        'priority' => _("Priority"),
                        'type' => _("Type"));
        foreach ($fields as $field => $label) {
            if ($name = $this->get($field . '_name')) {
                if (isset($this->_changes[$field])) {
                    $message .= '-' . String::pad($label, $length) . ' | ' . $this->_changes[$field]['from_name'] . "\n";
                    $message .= '+' . String::pad($label, $length) . ' | ' . $name . "\n";
                } else {
                    $message .= ' ' . String::pad($label, $length) . ' | ' . $name . "\n";
                }
            }
        }

        /* Show any change in ticket owners. */
        $owners = wordwrap(Whups::getOwners($this->_id, false, true), 72, "\n            | ");
        if (isset($this->_changes['oldowners'])) {
            $oldOwners = wordwrap(Whups::getOwners($this->_id, false, true, $this->_changes['oldowners']), 72, "\n            | ");
            $message .= '-' . String::pad(_("Owners"), $length) . ' | ' . $oldOwners . "\n";
            $message .= '+' . String::pad(_("Owners"), $length) . ' | ' . $owners . "\n";
        } else {
            $message .= ' ' . String::pad(_("Owners"), $length) . ' | ' . $owners . "\n";
        }

        /* Show any new files. */
        if (isset($this->_changes['attachment'])) {
            $message .= '+' . String::pad(_("Attachment"), $length) . ' | ' . $this->_changes['attachment']['to']['name'] . "\n";
        }

        /* Comment placeholder; will be filled in by Whups_Driver::mail(). */
        $message .= "-----------------------------------------------------------------------\n";
        $message .= "\n@@comment_placeholder@@";

        /* Add the "do not reply" tag if we don't monitor incoming  mail. */
        if (empty($conf['mail']['reply'])) {
            $message = $whups->splitMessage(_("DO NOT REPLY TO THIS MESSAGE. THIS EMAIL ADDRESS IS NOT MONITORED.")) . "\n\n" . $message;
        }

        /* Include NEW or RESOLVED in the subject for those states for easy
         * filtering/eyeballing. */
        if ($isNew) {
            $subject = sprintf(_("NEW: %s"), $this->get('summary'));
        } elseif ($whups->isCategory('resolved', $this->get('state'))) {
            $subject = sprintf(_("RESOLVED: %s"), $this->get('summary'));
        } else {
            $subject = $this->get('summary');
        }

        /* Pass off to Whups_Driver::mail() to do the actual comment fetching,
         * permissions checks, etc. */
        $whups->mail($this->_id, $whups->getListeners($this->_id, true, true, true),
                     $subject, $message, $author);
    }

}

/**
 * @package Whups
 */
class TicketDetailsForm extends Horde_Form {

    function TicketDetailsForm(&$vars, $title = '')
    {
        global $whups, $conf;

        parent::Horde_Form($vars, $title);

        $this->addVariable(_("Summary"), 'summary', 'text', true, true);
        $this->addVariable(_("Queue"), 'queue_name', 'text', true, true);

        if ($vars->get('version_name')) {
            $version = &$this->addVariable(_("Queue Version"), 'version', 'text', true, true);
            $version->setDefault($vars->get('version_name') . ': ' . $vars->get('version_description'));
        }

        $this->addVariable(_("Type"), 'type_name', 'text', true, true);
        $this->addVariable(_("State"), 'state_name', 'text', true, true);
        $this->addVariable(_("Priority"), 'priority_name', 'text', true, true);

        $owner = &$this->addVariable(_("Owners"), 'user_id_owner', 'email', true, true, null, array(false, true));
        $owner->setDefault(_("Unassigned"));

        $this->addVariable(_("Requester"), 'user_id_requester', 'email', true, true, null, array(false, true));
        $this->addVariable(_("Created"), 'timestamp', 'date', true, true);
        $this->addVariable(_("Assigned"), 'date_assigned', 'date', true, true);
        $this->addVariable(_("Resolved"), 'date_resolved', 'date', true, true);
        $this->addVariable(_("Attachments"), 'attachments', 'html', true, true);
    }

    function addAttributes($attribute_array)
    {
        $i = 0;

        foreach ($attribute_array as $name => $value) {
            $var = &$this->addVariable($name, '_attribute$i', 'text', true);
            $var->setDefault($value);
            $i++;
        }
    }

}
